"use strict";
Object.defineProperty(exports, "__esModule", {
    value: true
});
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        enumerable: true,
        get: all[name]
    });
}
_export(exports, {
    compileScopeOptions: ()=>compileScopeOptions,
    isConfigTuple: ()=>isConfigTuple,
    Scope: ()=>Scope,
    scope: ()=>scope,
    rootScope: ()=>rootScope,
    rootType: ()=>rootType,
    writeShallowCycleErrorMessage: ()=>writeShallowCycleErrorMessage,
    writeDuplicateAliasesMessage: ()=>writeDuplicateAliasesMessage
});
const _nodeTs = require("../nodes/node.js");
const _definitionTs = require("../parse/definition.js");
const _problemsTs = require("../traverse/problems.js");
const _chainableNoOpProxyTs = require("../utils/chainableNoOpProxy.js");
const _errorsTs = require("../utils/errors.js");
const _freezeTs = require("../utils/freeze.js");
const _pathsTs = require("../utils/paths.js");
const _cacheTs = require("./cache.js");
const _typeTs = require("./type.js");
function _checkPrivateRedeclaration(obj, privateCollection) {
    if (privateCollection.has(obj)) {
        throw new TypeError("Cannot initialize the same private elements twice on an object");
    }
}
function _classApplyDescriptorGet(receiver, descriptor) {
    if (descriptor.get) {
        return descriptor.get.call(receiver);
    }
    return descriptor.value;
}
function _classApplyDescriptorSet(receiver, descriptor, value) {
    if (descriptor.set) {
        descriptor.set.call(receiver, value);
    } else {
        if (!descriptor.writable) {
            throw new TypeError("attempted to set read only private field");
        }
        descriptor.value = value;
    }
}
function _classExtractFieldDescriptor(receiver, privateMap, action) {
    if (!privateMap.has(receiver)) {
        throw new TypeError("attempted to " + action + " private field on non-instance");
    }
    return privateMap.get(receiver);
}
function _classPrivateFieldGet(receiver, privateMap) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "get");
    return _classApplyDescriptorGet(receiver, descriptor);
}
function _classPrivateFieldInit(obj, privateMap, value) {
    _checkPrivateRedeclaration(obj, privateMap);
    privateMap.set(obj, value);
}
function _classPrivateFieldSet(receiver, privateMap, value) {
    var descriptor = _classExtractFieldDescriptor(receiver, privateMap, "set");
    _classApplyDescriptorSet(receiver, descriptor, value);
    return value;
}
function _classPrivateMethodGet(receiver, privateSet, fn) {
    if (!privateSet.has(receiver)) {
        throw new TypeError("attempted to get private field on non-instance");
    }
    return fn;
}
function _classPrivateMethodInit(obj, privateSet) {
    _checkPrivateRedeclaration(obj, privateSet);
    privateSet.add(obj);
}
var _ctx_type;
const compileScopeOptions = (opts)=>({
        codes: (0, _problemsTs.compileProblemWriters)(opts.codes),
        keys: opts.keys ?? "loose"
    });
let anonymousScopeCount = 0;
const scopeRegistry = {};
const spaceRegistry = {};
const isConfigTuple = (def)=>Array.isArray(def) && def[1] === ":";
var _resolutions = /*#__PURE__*/ new WeakMap(), _exports = /*#__PURE__*/ new WeakMap(), _register = /*#__PURE__*/ new WeakSet(), _cacheSpaces = /*#__PURE__*/ new WeakSet(), _initializeContext = /*#__PURE__*/ new WeakSet(), _resolveRecurse = /*#__PURE__*/ new WeakSet();
class Scope {
    getAnonymousTypeName(ctx) {
        let base = ctx ? ctx.path.length ? `${ctx.path}` : ctx.type.name : "type";
        if (base[0] !== "λ") {
            base = `λ${base}`;
        }
        let increment = 1;
        let name = base;
        while(this.isInternallyResolvable(name)){
            name = `${base}${++increment}`;
        }
        return name;
    }
    addAnonymous(type, ctx) {
        if (this.isInternallyResolvable(type.name)) {
            if (type === this.resolve(type.name)) {
                return type.name;
            }
            const name = this.getAnonymousTypeName(ctx);
            this.type([
                "node",
                type.node
            ], {
                name
            });
            return name;
        }
        _classPrivateFieldGet(this, _resolutions).set(type.name, type);
        return type.name;
    }
    get infer() {
        return _chainableNoOpProxyTs.chainableNoOpProxy;
    }
    compile() {
        if (!spaceRegistry[this.name]) {
            for(const name in this.aliases){
                this.resolve(name);
            }
            spaceRegistry[this.name] = _classPrivateFieldGet(this, _exports).root;
        }
        return _classPrivateFieldGet(this, _exports).root;
    }
    addReferenceIfResolvable(name, ctx) {
        const resolution = _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, name, "undefined", []);
        if (!resolution) {
            return false;
        }
        (_ctx_type = ctx.type).includesMorph || (_ctx_type.includesMorph = resolution.includesMorph);
        return true;
    }
    resolve(name) {
        return _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, name, "throw", []);
    }
    resolveNode(node) {
        return typeof node === "object" ? node : this.resolveNode(this.resolve(node).node);
    }
    isInternallyResolvable(name) {
        return _classPrivateFieldGet(this, _resolutions).has(name) || this.aliases[name];
    }
    constructor(aliases, opts = {}){
        _classPrivateMethodInit(this, _register);
        _classPrivateMethodInit(this, _cacheSpaces);
        _classPrivateMethodInit(this, _initializeContext);
        _classPrivateMethodInit(this, _resolveRecurse);
        _classPrivateFieldInit(this, _resolutions, {
            writable: true,
            value: void 0
        });
        _classPrivateFieldInit(this, _exports, {
            writable: true,
            value: void 0
        });
        this.aliases = aliases;
        this.parseCache = new _cacheTs.FreezingCache();
        _classPrivateFieldSet(this, _resolutions, new _cacheTs.Cache());
        _classPrivateFieldSet(this, _exports, new _cacheTs.Cache());
        this.expressions = {
            intersection: (l, r, opts)=>this.type([
                    l,
                    "&",
                    r
                ], opts),
            union: (l, r, opts)=>this.type([
                    l,
                    "|",
                    r
                ], opts),
            arrayOf: (def, opts)=>this.type([
                    def,
                    "[]"
                ], opts),
            keyOf: (def, opts)=>this.type([
                    "keyof",
                    def
                ], opts),
            node: (def, opts)=>this.type([
                    "node",
                    def
                ], opts),
            instanceOf: (def, opts)=>this.type([
                    "instanceof",
                    def
                ], opts),
            valueOf: (def, opts)=>this.type([
                    "===",
                    def
                ], opts),
            narrow: (def, fn, opts)=>this.type([
                    def,
                    "=>",
                    fn
                ], opts),
            morph: (def, fn, opts)=>this.type([
                    def,
                    "|>",
                    fn
                ], opts)
        };
        this.intersection = this.expressions.intersection;
        this.union = this.expressions.union;
        this.arrayOf = this.expressions.arrayOf;
        this.keyOf = this.expressions.keyOf;
        this.valueOf = this.expressions.valueOf;
        this.instanceOf = this.expressions.instanceOf;
        this.narrow = this.expressions.narrow;
        this.morph = this.expressions.morph;
        this.type = Object.assign((def, opts)=>{
            const name = opts?.name ? this.isInternallyResolvable(opts.name) ? (0, _errorsTs.throwParseError)(writeDuplicateAliasesMessage(opts.name)) : opts.name : this.getAnonymousTypeName();
            // remove name from opts since we've already used it and otherwise it will
            // interfere with traversal keys being ProblemOptions
            delete opts?.name;
            const t = (0, _typeTs.initializeType)(name, def, opts, this);
            const ctx = _classPrivateMethodGet(this, _initializeContext, initializeContext).call(this, t);
            t.node = (0, _freezeTs.deepFreeze)((0, _definitionTs.parseDefinition)(def, ctx));
            t.flat = (0, _freezeTs.deepFreeze)((0, _nodeTs.flattenType)(t));
            _classPrivateFieldGet(this, _resolutions).set(t.name, t);
            return t;
        }, {
            from: this.expressions.node
        });
        this.name = _classPrivateMethodGet(this, _register, register).call(this, opts);
        if (opts.standard !== false) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, [
                spaceRegistry["standard"]
            ], "imports");
        }
        if (opts.imports) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, opts.imports, "imports");
        }
        if (opts.includes) {
            _classPrivateMethodGet(this, _cacheSpaces, cacheSpaces).call(this, opts.includes, "includes");
        }
        this.config = compileScopeOptions(opts);
    }
}
function register(opts) {
    const name = opts.name ? scopeRegistry[opts.name] ? (0, _errorsTs.throwParseError)(`A scope named '${opts.name}' already exists`) : opts.name : `scope${++anonymousScopeCount}`;
    scopeRegistry[name] = this;
    return name;
}
function cacheSpaces(spaces, kind) {
    for (const space of spaces){
        for(const name in space){
            if (_classPrivateFieldGet(this, _resolutions).has(name) || name in this.aliases) {
                (0, _errorsTs.throwParseError)(writeDuplicateAliasesMessage(name));
            }
            _classPrivateFieldGet(this, _resolutions).set(name, space[name]);
            if (kind === "includes") {
                _classPrivateFieldGet(this, _exports).set(name, space[name]);
            }
        }
    }
}
function initializeContext(type) {
    return {
        type,
        path: new _pathsTs.Path()
    };
}
function resolveRecurse(name, onUnresolvable, seen) {
    const maybeCacheResult = _classPrivateFieldGet(this, _resolutions).get(name);
    if (maybeCacheResult) {
        return maybeCacheResult;
    }
    const aliasValue = this.aliases[name];
    if (!aliasValue) {
        return onUnresolvable === "throw" ? (0, _errorsTs.throwInternalError)(`Unexpectedly failed to resolve alias '${name}'`) : undefined;
    }
    const type = (0, _typeTs.initializeType)(name, aliasValue, undefined, this);
    _classPrivateFieldGet(this, _resolutions).set(name, type);
    _classPrivateFieldGet(this, _exports).set(name, type);
    const ctx = _classPrivateMethodGet(this, _initializeContext, initializeContext).call(this, type);
    const resolution = (0, _definitionTs.parseDefinition)(type.definition, ctx);
    if (typeof resolution === "string") {
        if (seen.includes(resolution)) {
            return (0, _errorsTs.throwParseError)(writeShallowCycleErrorMessage(name, seen));
        }
        seen.push(resolution);
        const resolvedType = _classPrivateMethodGet(this, _resolveRecurse, resolveRecurse).call(this, resolution, "throw", seen);
        type.node = resolvedType.node;
        type.flat = resolvedType.flat;
        return type;
    }
    type.node = (0, _freezeTs.deepFreeze)(resolution);
    type.flat = (0, _freezeTs.deepFreeze)((0, _nodeTs.flattenType)(type));
    return type;
}
const scope = (aliases, opts = {})=>new Scope(aliases, opts);
const rootScope = scope({}, {
    name: "empty",
    standard: false
});
const rootType = rootScope.type;
const writeShallowCycleErrorMessage = (name, seen)=>`Alias '${name}' has a shallow resolution cycle: ${[
        ...seen,
        name
    ].join("=>")}`;
const writeDuplicateAliasesMessage = (name)=>`Alias '${name}' is already defined`;
