import type { TraversalNode, TypeNode } from "../nodes/node.js";
import type { ParsedMorph } from "../parse/ast/morph.js";
import type { as, inferDefinition, validateDefinition } from "../parse/definition.js";
import type { ProblemOptions, Problems } from "../traverse/problems.js";
import type { defer, evaluate, xor } from "../utils/generics.js";
import type { BuiltinClass } from "../utils/objectKinds.js";
import type { Expressions } from "./expressions.js";
import type { Scope } from "./scope.js";
export type TypeParser<$> = {
    <def>(def: validateDefinition<def, $>): parseType<def, $>;
    <def>(def: validateDefinition<def, $>, opts: TypeOptions): parseType<def, $>;
} & TypeParserProps<$>;
export type TypeParserProps<$> = {
    from: Expressions<$>["node"];
};
export type parseType<def, $> = [def] extends [validateDefinition<def, $>] ? Type<inferDefinition<def, $>> : never;
type TypeRoot<t = unknown> = evaluate<{
    [as]: t;
    infer: asOut<t>;
    node: TypeNode;
    flat: TraversalNode;
    qualifiedName: QualifiedTypeName;
    definition: unknown;
    scope: Scope;
    includesMorph: boolean;
    config: TypeConfig | undefined;
}>;
export type KeyCheckKind = "loose" | "strict" | "distilled";
export type TypeOptions = evaluate<{
    keys?: KeyCheckKind;
    name?: string;
} & ProblemOptions>;
export type TypeConfig = Omit<TypeOptions, "name">;
export declare const initializeType: (name: string, definition: unknown, config: TypeConfig | undefined, scope: Scope) => Checker<unknown> & {
    infer: unknown;
    node: TypeNode;
    flat: TraversalNode;
    qualifiedName: QualifiedTypeName;
    definition: unknown;
    scope: Scope;
    includesMorph: boolean;
    config: TypeConfig | undefined;
    [as]: unknown;
};
export declare const isType: (value: unknown) => value is Type<unknown>;
export type CheckResult<t> = xor<{
    data: asOut<t>;
}, {
    problems: Problems;
}>;
type Checker<t> = (data: unknown) => CheckResult<t>;
export type Type<t = unknown> = defer<Checker<t> & TypeRoot<t>>;
export type QualifiedTypeName = `${string}.${string}`;
export type asIn<t> = asIo<t, "in">;
export type asOut<t> = asIo<t, "out">;
type asIo<t, io extends "in" | "out"> = t extends ParsedMorph<infer i, infer o> ? io extends "in" ? i : o : t extends object ? t extends BuiltinClass | ((...args: any[]) => any) ? t : {
    [k in keyof t]: asIo<t[k], io>;
} : t;
export {};
